/* studio device driver module */

/* Bug #1 : After severous driver error, reboot DSP is impossible.
*		Timing		: always
*		Reset		: electrical power off-on
*		Source bug	: unknow
*  Bug #2 : System hang-up when using a different DSP code than "b64"
*	(see ./include/studio.h)
*		Timing		: after several thounsands of interrupts
*		Reset		: reboot
*		Source bug	: unknow
*/

/*
 * Todo : linear monitor
 *	: param group control
 * 	: support for 2ADAT96
 * 	: Alsa mixer
 * 	: Alsa timer
 */

#define SND_MAIN_OBJECT_FILE
#include "studio.h"
#include "codec_studio.h"
#include "pcm_studio.h"
#include "st_switch.h"

/****************************************************************/
/* Module arguments:						*/
/*								*/
/* adat_sync: Source of ADAT sync for each card. No present or  */
/* -1 means CLK_PLL_OPTICAL_A            		        */
/* 	CLK_PLL_OPTICAL_A	0				*/
/*  	CLK_PLL_OPTICAL_B	1				*/
/*	CLK_PLL_EXTERNAL	2				*/
/*	CLK_44K			3				*/
/*	CLK_48K			4				*/
/*	CLK_PLL_TIMER1		5				*/
/*								*/
/* syncio: Optional Sync Backplate Card installed. No present   */
/* or -1 means no card.						*/
/*								*/
/* master_index: Index of the Master Studi/o soundcard. No      */
/* present or -1 means no master card. If master card present   */
/* there is only an Alsa-PCM subsystem manages all of codecs,   */
/* else every soundcard has its own PCM subsystem managing      */
/* its codec.	                                                */
/*								*/
/* monitor_on: Boolean flag (default, on) to manage monitor     */
/* outputs.                                                     */
/****************************************************************/

int	adat_sync[SND_CARDS]	= SND_DEFAULT_IDX;	/* Index 1-MAX */
int	syncio[SND_CARDS]	= SND_DEFAULT_IDX;	/* Index 1-MAX */
int	snd_index[SND_CARDS]	= SND_DEFAULT_IDX;	/* Index 1-MAX */
char	*snd_id[SND_CARDS]	= SND_DEFAULT_STR;	/* ID for this card */
int	master_index		= -1;
int	monitor_on		=  1;
#ifdef MODULE_PARM
MODULE_PARM(adat_sync, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(syncio, "Optional Sync Backplate Card installed");
MODULE_PARM(syncio_index, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(syncio_index, "Sync Backplate optional card installed");
MODULE_PARM(snd_index, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_index, "Index value for Studi/o soundcard.");
MODULE_PARM(snd_id, "1-" __MODULE_STRING(SND_CARDS) "s");
MODULE_PARM_DESC(snd_id, "ID string for Studi/o soundcard.");
MODULE_PARM(master_index, "i");
MODULE_PARM_DESC(master_index, "Index value for master soundcard.");
MODULE_PARM(monitor_on, "i");
MODULE_PARM_DESC(monitor_on, "Monitor analog outputs on/off");
#endif

struct snd_studio {
	struct pci_dev		*pci;
	snd_irq_t		*irqptr;

	snd_card_t		*card;
	int			registered;

	snd_pcm_studio_t	*pcm_s;
	codec_studio_t		*codec;
};

static struct snd_studio *snd_studio_cards[SND_CARDS] = SND_DEFAULT_PTR;


/****************************************************************/

static void snd_card_studio_use_inc(snd_card_t * card)
{
	MOD_INC_USE_COUNT;
}

/*************************************************/

static void snd_card_studio_use_dec(snd_card_t * card)
{
	MOD_DEC_USE_COUNT;
}

/*************************************************/

static void snd_card_studio_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct snd_studio *scard = (struct snd_studio *) dev_id;

	if (scard->pcm_s)
	if (scard->pcm_s->usage)
		snd_pcm_studio_interrupt(scard->pcm_s);
}

/*************************************************/
static int snd_card_studio_free(void)
{
	int			idx;
	struct snd_studio	*scard;

	/* Free switches */
	for (idx=0; idx<SND_CARDS; idx++){
		if (!(scard=snd_studio_cards[idx]))
			continue;
		if (scard->pcm_s)
			snd_studio_free_switches (
				scard->card, scard->pcm_s);
	}

	/* Free all */
	for (idx=0; idx<SND_CARDS; idx++){
		if (!(scard=snd_studio_cards[idx]))
			continue;

		if (scard->registered)
			snd_card_unregister(scard->card);

		if (scard->codec)
			snd_codec_studio_free(scard->codec);

		if (scard->pcm_s)
			snd_pcm_studio_free (scard->pcm_s);

		if (scard->card)
			snd_card_free(scard->card);

		snd_kfree(scard);
		snd_studio_cards[idx]=NULL;
	}

	return 0;
}

/*************************************************/
static int snd_card_studio_detect(unsigned *ncards)
{
	int			idx;
	struct snd_studio	*scard;
	struct pci_dev		*pci_d = NULL;

	/* Detect every Studi/o card */
	for (idx=0,*ncards=0; idx<SND_CARDS; idx++){
		if (!(pci_d = pci_find_device(
				PCI_VENDOR_ID_SONORUS,
				PCI_DEVICE_ID_SONORUS_STUDIO,
				pci_d)))
			break;
		if (!(scard = (struct snd_studio *)snd_kcalloc(sizeof(*scard), GFP_KERNEL)))
			goto __no_happy;
		scard->pci = pci_d;
		snd_studio_cards[(*ncards)++] = scard;
	}


	/* Is master index a correct card ? */
	if ((*ncards > 0) && (master_index >= 0))
	if (!snd_studio_cards[master_index]){
		snd_printk("Incorrect master index %i\n", master_index);
		goto __no_happy;
		}

	return 0;

      __no_happy:
	for (idx=0; idx<*ncards; idx++){
		snd_kfree(snd_studio_cards[idx]);
		snd_studio_cards[idx]=NULL;
	}
		
	return -ENXIO;
}

/*************************************************/

int init_module(void) {
	int			idx, icard, error;
	unsigned		ncards;
	struct snd_studio	*scard;
	snd_card_t		*card;
	snd_pcm_studio_t	*pcm_s;
	unsigned long		pci_base_address;

	/* Please, don't disturb */
	MOD_INC_USE_COUNT;

	/* Detect all cards */
	if ((error=snd_card_studio_detect(&ncards))){
		snd_printk("Problems at PCI detection\n");
		goto __bye;
		}


	/* Creating Alsa subsystems */
	for (idx=0; idx<SND_CARDS; idx++){
		if (!(scard=snd_studio_cards[idx]))
			continue;

		if (!(card = snd_card_new(snd_index[idx], snd_id[idx],
				snd_card_studio_use_inc,
				snd_card_studio_use_dec))){
			snd_printk("Problems creating Alsa-Studi/o card #%i\n", idx);
			error = -ENOMEM;
			goto __bye;
		}
		card->static_data	= scard;
		card->type		= SND_CARD_TYPE_STUDIO;
		scard->card		= card;
		scard->pcm_s		= NULL;
		scard->irqptr		= NULL;

		/* Create pcm subsystem only for master card, if present. */
		/* Else create for every card. Only cards with pcm        */
		/* subsystem need to register interrupt service.          */
		if ((master_index >=0) && (master_index != idx))
			continue;

		if (!(scard->pcm_s = snd_pcm_studio_create(card, idx,
					((master_index >= 0)) ? ncards : 1))){
			snd_printk("Problems creating PCM subsystem for Studi/o soundcard #%i\n", idx);
			error = -ENXIO;
			goto __bye;
		}

		if ((error = snd_register_interrupt(
				card, "Studi/o", scard->pci->irq,
				SND_IRQ_TYPE_PCI, snd_card_studio_interrupt,
				scard, NULL, &scard->irqptr)) < 0){
			snd_printk("Problems registering irq for Studi/o soundcard #%i\n", idx);
			goto __bye;
		}

	} /* for idx */

	/* Booting cards (creating codecs) */
	for (idx=0; idx<SND_CARDS; idx++){
		if (!(scard=snd_studio_cards[idx]))
			continue;

#ifdef NEW_PCI
		pci_base_address = scard->pci->resource[0].start;
#else
		pci_base_address = scard->pci->base_address[0]
					& ~PCI_BASE_ADDRESS_SPACE;
#endif

		if (!(scard->codec = snd_codec_studio_create(
				scard->card, idx,
				pci_base_address))){
			snd_printk("Problems booting Studi/o soundcard #%i\n", idx);
			error = -ENODEV;
			goto __bye;
			}
		/* Only master card has pcm subsystem         */
		/* If there is not master card every card has */
		/* its own pcm subsystem.                     */
		pcm_s = (master_index >= 0)
			? snd_studio_cards[master_index]->pcm_s
			: scard->pcm_s;
		pcm_s->codec[pcm_s->ncodecs++]= scard->codec;
		scard->codec->pcm_s = pcm_s;

	} /* for SND_CARDS */

	/* Creating switches */
	for (idx=0; idx<SND_CARDS; idx++){
		if (!(scard=snd_studio_cards[idx]))
			continue;

		/* Master card registers switches for all cards */
		/* If there is not master card every card has   */
		/* its own switch set.                          */
		card  = (master_index >= 0)
			? snd_studio_cards[master_index]->card
			: scard->card;
		icard = (master_index >= 0) ? master_index : idx;
		if (card == scard->card)
			if ((error=snd_studio_create_switches (
					card, icard,
					snd_studio_cards[icard]->pcm_s))){
				snd_printk("Problems creating switches for card #%i\n", icard);
				goto __bye;
			}
	} /* for SND_CARDS */


	/* Synchronize adat interfaces of codecs */
	if (master_index >= 0){
		pcm_s = snd_studio_cards[master_index]->pcm_s;
		if ((error=snd_codec_studio_sync_serial
				(pcm_s->codec, pcm_s->ncodecs))){
			snd_printk("Problems synchronizing codecs\n");
			goto __bye;
		}
	}

	/* Registering at Alsa home */
	for (idx=0; idx<SND_CARDS; idx++){
		if (!(scard=snd_studio_cards[idx]))
			continue;
		if (!(card=scard->card))
			continue;

#ifdef NEW_PCI
		pci_base_address = scard->pci->resource[0].start;
#else
		pci_base_address = scard->pci->base_address[0]
					& ~PCI_BASE_ADDRESS_SPACE;
#endif

		strcpy(card->abbreviation, "Studi/o");
		strcpy(card->shortname, "Sonorus Studi/o");
		sprintf(card->longname, "%s, at 0x%lx, irq %i.%s",
			card->shortname,
			pci_base_address,
			(scard->irqptr) ? scard->irqptr->irq : -1,
			(master_index == idx) ? " Master card" : "");

		if ((error = snd_card_register(scard->card))){
			snd_printk("Problems registering Alsa-Studi/o card #%i\n", idx);
			goto __bye;
		}
		scard->registered = 1;
	}

	/* Now ..., we are ready */
	MOD_DEC_USE_COUNT;

	return 0;

      __bye:
	snd_card_studio_free();
	MOD_DEC_USE_COUNT;
	return error;
}

/*************************************************/

void cleanup_module(void) {

	snd_card_studio_free();
}
