#ifndef _CODEC_STUDIO_H
#define _CODEC_STUDIO_H

/************/
/* DSP56301 PCI Memory Address Space, pag. 6-43 56301UM	*/
/*  REG : HCTR + HSTR + HCVR				*/
/* 	HCTR, HI32 Control Register			*/
/* 	HSTR, HI32 Status Register			*/
/* 	HCVR, Host Command Vector Register		*/
/*  DATA : HTXR/HRXS					*/
/* 	HTXR, Host Transmit Data Register		*/
/* 	HRXS, Slave Receive Data Register		*/
/************/
#define DSP56301_REG_OFFSET	0x0010
#define DSP56301_REG_SIZE	0x000C
#define DSP56301_DATA_OFFSET	0x001C
#define DSP56301_DATA_SIZE	0xFFE4

#define CONTROL_DSP(codec, l)		writel (l, (codec)->ba_reg)
#define STATUS_DSP(codec)		readl  ((codec)->ba_reg + 4)
#define COMMAND_DSP(codec, l)		writel (l, (codec)->ba_reg + 8)
#define DSP_U32_IN(codec)		readl  ((codec)->ba_data)
#define DSP_U32_OUT(codec, l)		writel (l, (codec)->ba_data)
#define DSP_DATA_OUT(codec, pl, nb)	memcpy_toio   ((codec)->ba_data, pl, nb)
#define DSP_ZEROED_OUT(codec, nb)	memset_io ((codec)->ba_data, 0, nb)
#define DSP_DATA_IN(codec, pl, nb)	memcpy_fromio (pl, (codec)->ba_data, nb)


/* HCTR, HI32 Control Register  */
#define HTF01	0x00000100L
#define HTF10	0x00000200L
#define HRF01	0x00000800L
#define HRF10	0x00001000L

/* HSTR, HI32 Status Register    */
/* 	H3-H5, Pag. 8-9, HISv0.9 */
#define TRDY			0x00000001L
#define TEST_TRDY(codec)	(STATUS_DSP(codec) & TRDY)
#define HTRQ			0x00000002L
#define TEST_HTRQ(codec)	(STATUS_DSP(codec) & HTRQ)
#define HRRQ			0x00000004L
#define TEST_HRRQ(codec)	(STATUS_DSP(codec) & HRRQ)
#define H3			0x00000008L
#define TEST_H3(codec)		((STATUS_DSP(codec) & H3) != 0L)
#define TEST_IDLE_STATE(codec)	(!TEST_H3(codec))
#define TEST_RUN_STATE(codec)	TEST_H3(codec)
#define H4			0x00000010L
#define TEST_H4(codec)		((STATUS_DSP(codec) & H4) != 0L)
#define TEST_DSP_READY(codec)	(!TEST_H4(codec))
#define TEST_DSP_BUSY(codec)	TEST_H4(codec)
#define H5			0x00000020L
#define TEST_H5(codec)		((STATUS_DSP(codec) & H5) != 0L)
#define TEST_CLK_LO(codec)	(!TEST_H5(codec))
#define TEST_CLK_HI(codec)	TEST_H5(codec)
#define HINT			0x00000040L
#define TEST_HINT(codec)	((STATUS_DSP(codec) & HINT) != 0L)

/* HCVR, Host Command Vector Register */
#define DSP56301_NMI		0x8000L



/************/
/* STUDI/O Host Interface Specification, HISv0.9 */
/************/

/* DSP56301 host command codes. Pag. 12, HISv0.9  */
/* Codes made up as (HC<<1)+1. Pag. 6-58, 56301UM */
#define HC_IDLE			0x000DL 
#define HC_RUN			0x000FL
#define HC_SEND_SETUP		0x003DL
#define HC_GET_SETUP		0x003FL
#define HC_SEND_STATUS		0x004DL
#define HC_GET_STATUS		0x004FL
#define HC_SEND_FPGABYTE	0x005BL 
#define HC_GET_FPGABYTE		0x005DL
#define HC_REPROG_FPGA		0x005FL
#define HC_SEND_PARAMS		0x0073L
#define HC_GET_PARAMS		0x0075L 
#define HC_STOP_SERIAL		0x0077L
#define HC_START_SERIAL		0x0079L
#define HC_ACK_HOSTINT		0x007BL 
#define HC_RESET_DSP		0x007DL
#define HC_ENABLE_W_INT		0x007FL
#define HC_DISABLE_W_INT	0x0081L
#define HC_ENABLE_R_INT		0x0083L
#define HC_DISABLE_R_INT	0x0085L
#define HC_READ_SAMPLES		0x0087L 
#define HC_WRITE_SAMPLES	0x0089L 
#define HC_WRITE_X_MEM		0x008BL
#define HC_WRITE_Y_MEM		0x008DL
#define HC_READ_X_MEM		0x008FL
#define HC_READ_Y_MEM		0x0091L
#define HC_SET_ADDRESS		0x0093L
#define HC_CLEAR_ERR		0x0095L 
#define HC_WR_EEPROM		0x0097L
#define HC_RD_EEPROM		0x0099L
#define HC_WR_MONITOR		0x009BL 
#define HC_HINT_SET		0x009DL
#define HC_READ_BLOCK_0		0x009FL
#define HC_READ_BLOCK_1		0x00A1L
#define HC_READ_BLOCK_2		0x00A3L
#define HC_READ_BLOCK_3		0x00A5L
#define HC_WRITE_BLOCK		0x00A7L
#define HC_READ_BLOCK		0x00A9L
 
#define HC_IDLE_NMI		(HC_IDLE 		| DSP56301_NMI)
#define HC_ACK_HOSTINT_NMI	(HC_ACK_HOSTINT		| DSP56301_NMI)
#define HC_SEND_STATUS_NMI	(HC_SEND_STATUS		| DSP56301_NMI)
#define HC_GET_STATUS_NMI	(HC_GET_STATUS		| DSP56301_NMI)
#define HC_SEND_PARAMS_NMI	(HC_SEND_PARAMS		| DSP56301_NMI)
#define HC_READ_SAMPLES_NMI	(HC_READ_SAMPLES	| DSP56301_NMI)
#define HC_WRITE_SAMPLES_NMI	(HC_WRITE_SAMPLES	| DSP56301_NMI)
#define HC_WR_MONITOR_NMI	(HC_WR_MONITOR		| DSP56301_NMI)
#define HC_CLEAR_ERR_NMI	(HC_CLEAR_ERR		| DSP56301_NMI)
#define HC_RESET_DSP_NMI	(HC_RESET_DSP		| DSP56301_NMI)
#define HC_READ_BLOCK_0_NMI	(HC_READ_BLOCK_0	| DSP56301_NMI)	
#define HC_READ_BLOCK_1_NMI	(HC_READ_BLOCK_1	| DSP56301_NMI)	
#define HC_READ_BLOCK_2_NMI	(HC_READ_BLOCK_2	| DSP56301_NMI)	
#define HC_READ_BLOCK_3_NMI	(HC_READ_BLOCK_3	| DSP56301_NMI)	
#define HC_WRITE_BLOCK_NMI	(HC_WRITE_BLOCK		| DSP56301_NMI)	
#define HC_READ_BLOCK_NMI	(HC_READ_BLOCK		| DSP56301_NMI)	

#define STUDIO_RETRY	150000L
#define ALTERA_RETRY	10000000L
#define TINY_DELAY	1200L

#define MODE_2ADAT	0
#define MODE_2ADAT96	5

/* Clock and PLL reference. Pag. 21, HISv0.9 */
#define CLK_PLL_OPTICAL_A	0x21L
#define CLK_PLL_OPTICAL_B	0x25L
#define CLK_PLL_EXTERNAL	0x2DL
#define CLK_44K			0x01L
#define CLK_48K			0x11L
#define CLK_PLL_TIMER1		0x29L

typedef struct snd_codec_studio codec_studio_t;
struct snd_codec_studio {
	snd_card_t		*card;
	void			*pcm_s;
	unsigned		icodec;
	int			syncio;		/* Optional Sync Backplate Card installed */
	int			adat_sync;	/* ADAT sync type */

	spinlock_t		access_lock;
	u16			bitmap_read;	/* Managed at pcm layer */

        void			*ba_reg;	/* Virtual Base Addres for registers */
        void			*ba_data;	/* Virtual Base Addres for data */

	int                      running;	/* 0 : Idle, <>0 : Run */
	u32			 cookie[2];     /* Pag. 2, HISv0.9 */
	struct {
		u32	error;
#define UP_REV_D	0x000040L
#define GENERIC_ERROR	0x200000L
#define DRAM_ERROR	0x400000L
#define RT_INT_ERROR	0x800000L
		u32	ltime;
		u32	iostat;
#define NO_ADAT_SYNC		0x00000001L
#define INPUT_A_DATA_ERROR	0x00000002L
#define INPUT_B_DATA_ERROR	0x00000004L
		u32	reserve;		/* debugging    */
		u32	sample_rate_measured;
		u32	pingpong_addr;		/* debugging    */
		u32	sample_counter_lo;	/* low 24 bits  */
		u32	sample_counter_hi;	/* high 24 bits */
		u32	timecode_lo;		/* low 24 bits  */
		u32	timecode_hi;		/* high 24 bits */
		u32	peak_playback[16];	/* peak per frame */
		u32	peak_record[16];
		u32	peak_monitor[2];
	} status; /* Pag. 8-9, HISv0.9 */
	int		up_rev_d;
	struct {
		u32	format;
#define TCEN	0x200L
		u32	nplay;
		u32	nrecord;
		u32	mon1l;
		u32	mon2l;
		u32	mon1r;
		u32	mon2r;
		u32	scale_factor;	/* for meters */
		u32	nsamples;	/* per channel */
		u32	rate_divider;
		u32	reserve[6];
	} setup; /* Pag. 6-7, HISv0.9, "reserve" field is undocumented */
	struct {
		u32	output_monitor_level[32];
		u32	master_monitor_level[2];
		u32	input_monitor_level[32];
		u32	DA_control;
		u32	output_phases;
		u32	input_phases;
	} params; /* Pag. 10-11, HISv0.9 */
	struct {
		unsigned long	nretries_dsp_busy;
		unsigned long	sum_retries_dsp_busy;
		unsigned long	nretries_dsp_ready;
		unsigned long	sum_retries_dsp_ready;
		unsigned long	nretries_not_htrq;
		unsigned long	sum_retries_not_htrq;
		unsigned long	nretries_not_hrrq;
		unsigned long	sum_retries_not_hrrq;
		unsigned long	nretries_not_trdy;
		unsigned long	sum_retries_not_trdy;
		unsigned long	nretries_not_run;
		unsigned long	sum_retries_not_run;
		unsigned long	nretries_not_idle;
		unsigned long	sum_retries_not_idle;
		unsigned long	sum_wait_uS;
	} stat;

	/* Routing data are local to codec. However,     */
	/* it is managed to pcm layer. Switches are      */
	/* hosted in control switch list (see st_switch.c) */
	snd_kswitch_t	*sw_playback_routing;
	unsigned short	playback_routing[16];
	snd_kswitch_t	*sw_record_routing;
	unsigned short	record_routing[16];
	snd_kswitch_t	*sw_monitor_routing;
	unsigned short	monitor_routing[2];

	snd_kswitch_t	*sw_playback0_meters;
	snd_kswitch_t	*sw_playback1_meters;
	snd_kswitch_t	*sw_record0_meters;
	snd_kswitch_t	*sw_record1_meters;
	snd_kswitch_t	*sw_monitor_meters;

	snd_info_entry_t *entry_proc;

}; /* struct snd_codec_studio */

#define TEST_RT_ERROR(codec)		((codec)->status.error  & RT_INT_ERROR)
#define CLEAR_RT_ERROR(codec)		(codec)->status.error  &= ~RT_INT_ERROR
#define TEST_ADAT_SYNC_ERROR(codec)	((codec)->status.iostat & NO_ADAT_SYNC)

unsigned int snd_codec_get_channel_size(void);


/* buffer must be 16*CHANNEL_NBYTES-sized or NULL */
int snd_codec_studio_playback(
		codec_studio_t *codec,
		void *buffer);
/* buffer must be 16*CHANNEL_NBYTES-sized or NULL */
int snd_codec_studio_record(
		codec_studio_t *codec,
		void *buffer);
/* buffer must be 2*CHANNEL_NBYTES-sized or NULL */
int snd_codec_studio_monitor(
		codec_studio_t *codec,
		void *buffer);
int snd_codec_studio_status(
		codec_studio_t *codec);
int snd_codec_studio_clear_error(
	codec_studio_t	**lcodec,
	unsigned	ncodec);
int snd_codec_studio_test_int(
		codec_studio_t *codec);
int snd_codec_studio_ackint(
		codec_studio_t *codec);


int snd_codec_studio_start(
	codec_studio_t	**lcodec,
	unsigned	ncodec);
int snd_codec_studio_stop(
	codec_studio_t	**lcodec,
	unsigned	ncodec);
int snd_codec_studio_sync_serial(
	codec_studio_t	**lcodec,
	unsigned	ncodec);
int snd_codec_studio_test_routes(
	codec_studio_t * codec);
void snd_codec_studio_free (
	codec_studio_t *codec);
codec_studio_t *snd_codec_studio_create (
	snd_card_t	*card,
	unsigned	icodec,
	unsigned long	pci_base_address);

#endif /* _CODEC_STUDIO_H */
