#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <sys/asoundlib.h>
#include <sys/ioctl.h>


#define MAX_CH		32
#define HD_OP		0

static snd_ctl_t	*handle_ctl	= NULL;
static int		nfds		= 0;
static snd_pcm_t	*handle_pcm[MAX_CH];
static int		fd_pcm[MAX_CH];
static void		*bufin_pcm[MAX_CH];
static unsigned		sizein_pcm[MAX_CH];
static unsigned		num_pcm_subch = 0;

#if HD_OP
#define DIR_AUDIO	"/home/audio/db/test"
static int		fd_write[MAX_CH];
static char		bufstr[512];
#endif

/*****************************************************************/
static void alsa_free (void)
{
	unsigned idx;

	for (idx=0; idx<num_pcm_subch; idx++){
#if HD_OP
		if (fd_write[idx])
    			close(fd_write[idx]);
#endif
		if (handle_pcm[idx])
    			snd_pcm_close(handle_pcm[idx]);
		if (bufin_pcm[idx])
			free(bufin_pcm[idx]);
	}
}

/*****************************************************************/
static void alsa_error (char *msg, int ierror)
{
	fprintf (stderr, "test.c: %s: %s\n", msg, snd_strerror(ierror));

	alsa_free();

	exit(1);
}

/*****************************************************************/
static int test_control (snd_ctl_t *hdl, int icard)
{
	int			ierror, isw;
	snd_switch_list_item_t  pswitches[50];
	snd_switch_list_t	lsw;
	snd_switch_t		sw;

	/* List of control switches. See st_switch.c */
	lsw.switches_size = 50;
	lsw.pswitches     = pswitches;
	if ((ierror = snd_ctl_switch_list(hdl, &lsw))<0)
   		alsa_error ("test_control: snd_ctl_switch_list", ierror);
	for (isw=0; isw<lsw.switches; isw++){
		sprintf (sw.name, "%s", lsw.pswitches[isw].name);
		if ((ierror=snd_ctl_switch_read(hdl, &sw))<0)
   			alsa_error ("test_control: snd_ctl_switch_read", ierror);

		fprintf (stderr, "Control Switch %d: %s\n", isw, sw.name);
	} /* for isw */
#if 1
	/* Routing of record channel #4 to playback channels #4,#5 */
	memset (&sw, 0, sizeof(sw));
	sprintf (sw.name, "Play Routes (1-16) of Card#%d", icard);
	sw.type = SND_SW_TYPE_WORD;
	sw.value.data16[3] = 16*1+4;	/* Record #4 --> Playback #4 */
	sw.value.data16[4] = 16*1+4;	/* Record #4 --> Playback #5 */
	if ((ierror=snd_ctl_switch_write(hdl, &sw))<0)
   		alsa_error ("test_control: snd_ctl_switch_write", ierror);
#endif

#if 1
	/* Routing of playback channel #1 to monitor channel #1 */
	/* Routing of playback channel #2 to monitor channel #2 */
	memset (&sw, 0, sizeof(sw));
	sprintf (sw.name, "Mon Routes (1-2) of Card#%d", icard);
	sw.type = SND_SW_TYPE_WORD;
	sw.value.data16[0] = 1;
	sw.value.data16[1] = 2;
	if ((ierror=snd_ctl_switch_write(hdl, &sw))<0)
   		alsa_error ("test_control: snd_ctl_switch_write", ierror);
#endif

	return(1);
}

/*****************************************************************/
static int info_pcm (snd_pcm_t *hdl)
{
	int                     ierror;
	snd_pcm_info_t          pcm_info;
	snd_pcm_playback_info_t pcm_playback_info;
	snd_pcm_capture_info_t   pcm_record_info;

	if ((ierror=snd_pcm_info(hdl, &pcm_info)) < 0)
   		alsa_error ("info_pcm: snd_pcm_info", ierror);
	if (!(pcm_info.flags & SND_PCM_INFO_PLAYBACK))
   		fprintf (stderr, "test.c: PCM_INFO_PLAYBACK no available\n");
	if (!(pcm_info.flags & SND_PCM_INFO_CAPTURE))
		fprintf (stderr, "test.c: PCM_INFO_CAPTURE no available\n");
   
	if ((ierror=snd_pcm_playback_info(hdl, &pcm_playback_info)) < 0)
		alsa_error ("info_pcm: snd_pcm_playback_info", ierror);

	fprintf (stderr, "test.c: Fragment size playback [%u, %u]\n",
		pcm_playback_info.min_fragment_size,
		pcm_playback_info.max_fragment_size);


	if (pcm_playback_info.formats & SND_PCM_FMT_S32_LE)
   		fprintf (stderr, "test.c: PCMp S32_LE\n");
	if (pcm_playback_info.formats & SND_PCM_FMT_S32_BE)
   		fprintf (stderr, "test.c: PCMp S32_BE\n");
	if (pcm_playback_info.formats & SND_PCM_FMT_U32_LE)
   		fprintf (stderr, "test.c: PCMp U32_LE\n");
	if (pcm_playback_info.formats & SND_PCM_FMT_U32_BE)
   		fprintf (stderr, "test.c: PCMp U32_BE\n");

	if ((ierror=snd_pcm_capture_info(hdl, &pcm_record_info)) < 0)
		alsa_error ("info_pcm: snd_pcm_record_info", ierror);

	fprintf (stderr, "test.c: Fragment size record [%u, %u]\n",
		pcm_record_info.min_fragment_size,
		pcm_record_info.max_fragment_size);

	if (pcm_record_info.formats & SND_PCM_FMT_S32_LE)
   		fprintf (stderr, "test.c: PCMr S32_LE\n");
	if (pcm_record_info.formats & SND_PCM_FMT_S32_BE)
   		fprintf (stderr, "test.c: PCMr S32_BE\n");
	if (pcm_record_info.formats & SND_PCM_FMT_U32_LE)
   		fprintf (stderr, "test.c: PCMr U32_LE\n");
	if (pcm_record_info.formats & SND_PCM_FMT_U32_BE)
   		fprintf (stderr, "test.c: PCMr U32_BE\n");

	return(1);
}

/*****************************************************************/
static snd_pcm_t *new_pcm_subchannel (
			int icard, int idx)
{
	int				ierror;
	snd_pcm_playback_status_t	pcm_playback_status;
	snd_pcm_capture_status_t	pcm_record_status;
	snd_pcm_playback_params_t	pcm_playback_params;
	snd_pcm_capture_params_t	pcm_record_params;
	snd_pcm_t			*hdl;
	snd_pcm_format_t		format;

	/* Open Alsa subchannel read/write */
	if ((ierror=snd_pcm_open_subdevice(
			&handle_pcm[idx], icard,
			0, idx, SND_PCM_OPEN_DUPLEX)) < 0)
		return ((snd_pcm_t *)NULL);
	hdl = handle_pcm[idx];

	if ((fd_pcm[idx] = snd_pcm_file_descriptor(hdl)) < 0)
   		alsa_error ("snd_pcm_file_descriptor", fd_pcm[idx]);
	if (nfds <= fd_pcm[idx])
		nfds = fd_pcm[idx]+1;

#if HD_OP
	/* Open local file to write */
	sprintf (bufstr, "%s/test%d", DIR_AUDIO, idx);
	if ((fd_write[idx] = open(bufstr, O_WRONLY|O_CREAT|O_TRUNC,
					0777))<0)
   		alsa_error ("open", fd_write[idx]);
#endif

	/* Set sample rate */
	format.format   = SND_PCM_SFMT_U32_LE;
	format.rate     = 48000;
	format.channels = 1;
	format.special  = 0;
	if ((ierror = snd_pcm_playback_format(hdl, &format))<0)
   		alsa_error ("snd_pcm_playback_format", ierror);
	if ((ierror = snd_pcm_capture_format(hdl, &format))<0)
   		alsa_error ("snd_pcm_capture_format", ierror);

	/* Blocking mode. You had better avoid non-blocking */
	/* mode, see ./card_studio/studio.c                 */
	if ((ierror=snd_pcm_block_mode(hdl, 1)) < 0)
   		alsa_error ("set_pcm: snd_pcm_block_mode", ierror);

	if (snd_pcm_playback_status(hdl, &pcm_playback_status))
   		alsa_error ("set_pcm: snd_pcm_playback_status", ierror);
	if (snd_pcm_capture_status(hdl, &pcm_record_status))
   		alsa_error ("set_pcm: snd_pcm_record_status", ierror);

	/* Buffer things. Actually, fragment size is forced by driver */
	pcm_playback_params.fragment_size   = pcm_playback_status.fragment_size;
	pcm_playback_params.fragments_max   = pcm_playback_status.fragments;
	pcm_playback_params.fragments_room  = 1;
	pcm_record_params.fragment_size     = pcm_record_status.fragment_size;
	pcm_record_params.fragments_min     = 2;
	if ((ierror=snd_pcm_playback_params(hdl, &pcm_playback_params)) < 0) 
   		alsa_error ("set_pcm: snd_pcm_playback_params", ierror);
	if ((ierror=snd_pcm_capture_params(hdl, &pcm_record_params)) < 0) 
   		alsa_error ("set_pcm: snd_pcm_record_params",ierror);

	/* Local buffer */
	sizein_pcm[idx]  = pcm_record_status.fragment_size;
	sizein_pcm[idx] *= pcm_record_params.fragments_min;
	bufin_pcm[idx]   = malloc(sizein_pcm[idx]);

	/* Playing initialization */
	if ((ierror=snd_pcm_drain_playback(hdl)) < 0)
   		alsa_error ("set_pcm: snd_pcm_drain_playback", ierror);
	/* Recording initialization */
	if ((ierror=snd_pcm_flush_capture(hdl)) < 0)
   		alsa_error ("set_pcm: snd_pcm_flush_capture", ierror);

	return hdl;
}

/*****************************************************************/
int main (int  argc, char *argv[])
{
	int				ncards, icard, iter;
	int				idx, ierror;
	ssize_t				nbytes;
	fd_set				readfs, writefs, exceptfs;
	snd_pcm_playback_status_t	pcm_playback_status;
	snd_pcm_capture_status_t	pcm_record_status;

	/* Safe */
	sync ();
iter = 0;

	memset (handle_pcm, 0, sizeof(handle_pcm));
	memset (fd_pcm,     0, sizeof(fd_pcm));
	memset (bufin_pcm,  0, sizeof(bufin_pcm));
	memset (sizein_pcm, 0, sizeof(sizein_pcm));

#if HD_OP
	memset (fd_write,   0, sizeof(fd_write));
#endif

	/* Test for loaded modules */
	if ((ncards = snd_cards()) < 0)
   		alsa_error ("snd_cards", ncards);
	if (ncards == 0){
   		fprintf (stderr, "test.c: Sound subsystem not avaliable\n");
   		exit(1);
   	}

	/* Get card by name */
	if ((icard = snd_card_name("studio0")) < 0)
   		alsa_error ("snd_card_name", ncards);

	/* Open digital audio subsystem, all channels */
	for (num_pcm_subch=0; num_pcm_subch</*MAX_CH*/20; num_pcm_subch++)
		if (!new_pcm_subchannel(icard, num_pcm_subch))
			break;
	if (!num_pcm_subch){
   		fprintf (stderr, "test.c: No avaliable channels\n");
		return 1;
	}
   	fprintf (stderr, "test.c: #channels: %u\n", num_pcm_subch);

	/* Get info digital audio, use channel #0*/
	info_pcm (handle_pcm[0]);

	/* Safe */
	sync ();

	/* Open control subsystem */
	if ((ierror=snd_ctl_open(&handle_ctl, icard)) < 0)
   		alsa_error ("snd_ctl_open", ierror);

	/* Safe */
	sync ();
	test_control (handle_ctl, icard);

	/* Safe */
	sync ();

	/* infinite loop */
	for (;;){
		FD_ZERO(&writefs);
		FD_ZERO(&readfs);
		FD_ZERO(&exceptfs);

		for (idx=0; idx<num_pcm_subch; idx++){
			FD_SET(fd_pcm[idx], &writefs);
			FD_SET(fd_pcm[idx], &readfs);
		}

    		/* Select */
    		while (select (nfds, &readfs, &writefs, &exceptfs, NULL) == -1)
			if (errno != EINTR){
          			fprintf (stderr, "test.c: error select\n");
          			exit(1);
			}

#if 1
    		/* Capture */
		for (idx=0; idx<num_pcm_subch; idx++)
		if (FD_ISSET(fd_pcm[idx], &readfs)){
    			if ((nbytes = snd_pcm_read(handle_pcm[idx],
					bufin_pcm[idx], sizein_pcm[idx])) < 0)
				fprintf (stderr, "test.c: snd_pcm_read error\n");

			/* Test Overrun */
			if ((ierror=snd_pcm_capture_status(handle_pcm[idx], &pcm_record_status)) < 0)
				alsa_error ("snd_pcm_record_status", ierror);
			if (pcm_record_status.overrun)
				fprintf (stderr, "test.c: overrun channel #%d\n", idx+1);

#if HD_OP
			/* Write to disk */
			if ((write(fd_write[idx], bufin_pcm[idx],
					sizein_pcm[idx]) != sizein_pcm[idx]))
				fprintf (stderr, "test.c: Fail HD recording channel #%d\n", idx+1);
#endif
		}
#endif


#if 1
    		/* Playback */
		for (idx=0; idx<num_pcm_subch; idx++)
		if (FD_ISSET(fd_pcm[idx], &writefs)){
    			if ((nbytes = snd_pcm_write(handle_pcm[idx],
					bufin_pcm[idx], sizein_pcm[idx])) < 0)
				fprintf (stderr, "test.c: snd_pcm_write error\n");
			/* Test Underrun */
			if ((ierror=snd_pcm_playback_status(handle_pcm[idx], &pcm_playback_status)) < 0)
				alsa_error ("snd_pcm_playback_status", ierror);
			if (pcm_playback_status.underrun)
				fprintf (stderr, "test.c: underrun channel #%d\n", idx+1);
			}
#endif
		iter++;
		if ((iter%2000)==0)
         		fprintf (stderr, "test.c: %d\n", iter);
		} /* for ;; */

	return(0);
}
